"""Client for connecting to SETRA API"""
import logging
import urllib.parse
from datetime import date
from typing import Union, Optional, List

import requests

from setra_client.models import (
    BatchErrors,
    BatchProgressEnum,
    CompleteBatch,
    InputBatch,
    OutputBatch,
    Parameter,
    Order,
    Detail,
    AbwOrder,
    AbwOrderErrors,
)

logger = logging.getLogger(__name__)


def merge_dicts(*dicts):
    """
    Combine a series of dicts without mutating any of them.

    >>> merge_dicts({'a': 1}, {'b': 2})
    {'a': 1, 'b': 2}
    >>> merge_dicts({'a': 1}, {'a': 2})
    {'a': 2}
    >>> merge_dicts(None, None, None)
    {}
    """
    combined = dict()
    for d in dicts:
        if not d:
            continue
        for k in d:
            combined[k] = d[k]
    return combined


class SetraEndpoints:
    def __init__(self,
                 url,
                 batch_url='api/batch/',
                 transaction_url='api/transaction/',
                 voucher_url='api/voucher/',
                 new_batch_url='api/addtrans/',
                 put_batch_url='api/addtrans/',
                 batch_complete_url='api/batch_complete/',
                 batch_error_url="api/batch_error/",
                 parameters_url="api/parameters/",

                 # Order urls (sotra):
                 order_url="api/order/",
                 order_complete_url="api/order_complete/",
                 detail_url="api/detail/",
                 details_in_order_url="api/details_in_order/",
                 abw_order_complete_url="api/abw_order_complete/",
                 post_add_abw_order_url="api/add_abw_order/",
                 abw_order_errors_url="api/abw_order_errors/",
                 ):

        self.baseurl = url
        self.batch_url = batch_url
        self.transaction_url = transaction_url
        self.voucher_url = voucher_url
        self.new_batch_url = new_batch_url
        self.put_batch_url = put_batch_url
        self.batch_complete_url = batch_complete_url
        self.batch_error_url = batch_error_url
        self.parameters_url = parameters_url
        # order urls:
        self.order_url = order_url
        self.order_complete_url = order_complete_url
        self.detail_url = detail_url
        self.details_in_order_url = details_in_order_url
        self.abw_order_complete_url = abw_order_complete_url
        self.post_add_abw_order_url = post_add_abw_order_url
        self.abw_order_errors_url = abw_order_errors_url

    """ Get endpoints relative to the SETRA API URL. """

    def __repr__(self):
        return '{cls.__name__}({url!r})'.format(
            cls=type(self),
            url=self.baseurl)

    def batch(self, batch_id: str = None):
        """
        URL for Batch endpoint
        """
        if batch_id is None:
            return urllib.parse.urljoin(self.baseurl, self.batch_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.batch_url, batch_id)))

    def transaction(self, trans_id: str = None):
        """
        Url for Transaction endpoint
        """
        if trans_id is None:
            return urllib.parse.urljoin(self.baseurl, self.transaction_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.transaction_url,
                                                  trans_id)))

    def voucher(self, vouch_id: str = None):
        """
        Url for Voucher endpoint
        """
        if vouch_id is None:
            return urllib.parse.urljoin(self.baseurl, self.voucher_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.voucher_url, vouch_id)))

    def post_new_batch(self):
        return urllib.parse.urljoin(self.baseurl, self.new_batch_url)

    def put_update_batch(self):
        return urllib.parse.urljoin(self.baseurl, self.new_batch_url)

    def batch_complete(self, batch_id: str):
        """
        URL for Batch endpoint
        """
        return urllib.parse.urljoin(self.baseurl,
                                    '/'.join((self.batch_complete_url, batch_id)))

    def batch_error(self, batch_id: str):
        """
        URL for batch_error endpoint
        """
        return urllib.parse.urljoin(
            self.baseurl, '/'.join((self.batch_error_url, batch_id))
        )

    def parameters(self):
        """Get url for parameters endpoint"""
        return urllib.parse.urljoin(self.baseurl, self.parameters_url)

    def abw_order_complete(self, abw_order_id: str = None):
        """
        URL for getting an abw order including a list of its orders, and each order contains a list of its detail objects
        """
        return urllib.parse.urljoin(self.baseurl,
                                    '/'.join((self.abw_order_complete_url, abw_order_id)))

    def add_abw_order(self):
        """
        URL for posting a new complete abw order containing order and detail objects
        """
        return urllib.parse.urljoin(self.baseurl, self.post_add_abw_order_url)

    def order(self, order_id: str = None):
        """
        URL for order endpoint
        """
        if order_id is None:
            return urllib.parse.urljoin(self.baseurl, self.order_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.order_url, order_id)))

    def order_complete(self, order_id: str):
        """
        URL for getting an order object including a list of its detail objects
        """
        return urllib.parse.urljoin(self.baseurl,
                                    '/'.join((self.order_complete_url, order_id)))

    def detail(self, detail_id: str):
        """
        URL for detail endpoint
        """
        return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.detail_url, detail_id)))

    def details_in_order(self, order_id: str):
        """
        URL for getting a list of detail objects in an order
        """
        return urllib.parse.urljoin(self.baseurl,
                                    '/'.join((self.details_in_order_url, order_id)))

    def abw_order_errors(self, abw_order_id: str):
        """
        URL for getting an object containing lists of all errors for AbwOrder, Orders and Details
        """
        return urllib.parse.urljoin(self.baseurl,
                                    '/'.join((self.abw_order_errors_url, abw_order_id)))


class SetraClient(object):
    default_headers = {
        'Accept': 'application/json',
    }

    def __init__(self,
                 url: str,
                 headers: Union[None, dict] = None,
                 return_objects: bool = True,
                 use_sessions: bool = True,
                 ):
        """
        SETRA API client.

        :param str url: Base API URL
        :param dict headers: Append extra headers to all requests
        :param bool return_objects: Return objects instead of raw JSON
        :param bool use_sessions: Keep HTTP connections alive (default True)
        """

        self.urls = SetraEndpoints(url)
        self.headers = merge_dicts(self.default_headers, headers)
        self.return_objects = return_objects
        if use_sessions:
            self.session = requests.Session()
        else:
            self.session = requests

    def _build_request_headers(self, headers):
        request_headers = {}
        for h in self.headers:
            request_headers[h] = self.headers[h]
        for h in (headers or ()):
            request_headers[h] = headers[h]
        return request_headers

    def call(self,
             method_name,
             url,
             headers=None,
             params=None,
             return_response=True,
             **kwargs):
        headers = self._build_request_headers(headers)
        if params is None:
            params = {}
        logger.debug('Calling %s %s with params=%r',
                     method_name,
                     urllib.parse.urlparse(url).path,
                     params)
        r = self.session.request(method_name,
                                 url,
                                 headers=headers,
                                 params=params,
                                 **kwargs)
        if r.status_code in (500, 400, 401):
            logger.warning('Got HTTP %d: %r', r.status_code, r.content)
        if return_response:
            return r
        r.raise_for_status()
        return r.json()  # Note: krasjer her, dersom man får text og ikke json i requesten, og return_response=false

    def get(self, url, **kwargs):
        return self.call('GET', url, **kwargs)

    def put(self, url, **kwargs):
        return self.call('PUT', url, **kwargs)

    def post(self, url, **kwargs):
        return self.call('POST', url, **kwargs)

    def object_or_data(
            self, cls, data: Union[dict, List[dict]]
    ) -> Union[object, dict]:
        """Create list of objects or return data as is"""
        if not self.return_objects:
            return data
        if isinstance(data, dict):
            return cls.from_dict(data)
        if isinstance(data, list):
            return [cls.from_dict(i) for i in data]
        raise TypeError(f"Expected list or dict, got {type(data)}")

    def search_batches(
        self,
        min_created_date: Optional[date] = None,
        max_created_date: Optional[date] = None,
        batch_progress: Optional[BatchProgressEnum] = None,
        interface: Optional[str] = None,
    ) -> Union[List[dict], dict, List[OutputBatch]]:
        """
        Search batches in SETRA.
        Dates (maximal and minimal creation dates) should
        be defined like ISO strings - like "2020-01-01".
        Batch Progress and interface - like strings exactly
        as the values that can be found in the corresponding
        fields of Setra.
        """

        params = {
            'min_created_date': min_created_date,
            'max_created_date': max_created_date,
            'batch_progress': batch_progress,
            'interface': interface,
        }

        url = self.urls.batch()
        response = self.get(url, params=params)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            if isinstance(data, list):
                return [OutputBatch(**item) for item in data]
            elif isinstance(data, dict):
                return [OutputBatch(**data)]
        else:
            return data

    def get_batch(self, batch_id: str) -> Union[OutputBatch, dict]:
        """
        GETs a batch from SETRA.
        """

        url = self.urls.batch(str(batch_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            return OutputBatch(**data)
        else:
            return data

    def get_voucher(self, vouch_id: int = None):
        """
        GETs one or all vouchers from SETRA
        """
        if vouch_id is not None:
            vouch_id = str(vouch_id)

        url = self.urls.voucher(vouch_id)
        response = self.get(url)
        return response.json()

    def get_transaction(self, trans_id: int = None):
        """
        GETs one or all transactions from SETRA
        """
        if trans_id is not None:
            trans_id = str(trans_id)

        url = self.urls.transaction(trans_id)
        response = self.get(url)
        return response.json()

    def post_new_batch(self, batchdata: InputBatch):
        """
        POST combination of batch, vouchers and transactions
        """
        url = self.urls.post_new_batch()
        headers = {'Content-Type': 'application/json'}
        response = self.post(url,
                             data=batchdata.json(exclude_unset=True),
                             headers=headers,
                             return_response=True)
        try:
            content = response.json()
        except requests.exceptions.JSONDecodeError:
            content = response.content

        if response.status_code == 202:
            return 'Accepted', {}
        elif response.status_code == 409:
            return 'Conflict', content
        else:
            return 'Unknown', {'code': response.status_code, 'content': content}

    def put_update_batch(self, batchdata: InputBatch):
        """
        PUT updates an existing batch with vouchers and transactions,
        if the batch exists in setra, and has status=created, or validation failed.
        Returns 404 if no batch was found, 204 for successful update, or 409 if batch
        was found, but did not meet the status criteria mentioned above.
        """
        url = self.urls.put_update_batch()
        headers = {'Content-Type': 'application/json'}
        response = self.put(url,
                            data=batchdata.json(exclude_unset=True),
                            headers=headers,
                            return_response=True)

        try:
            content = response.json()
        except requests.exceptions.JSONDecodeError:
            content = response.content

        if response.status_code == 204:
            return 'Accepted', {}
        elif response.status_code == 409:
            return 'Conflict', content
        else:
            return 'Unknown', {'code': response.status_code, 'content': content}

    def get_batch_complete(self, batch_id: str):
        """
        GETs complete batch (with vouchers and transactions)
        from SETRA
        """
        url = self.urls.batch_complete(batch_id)
        response = self.get(url)
        return self.object_or_data(CompleteBatch, response.json())

    def get_batch_errors(self, batch_id: str):
        url = self.urls.batch_error(batch_id)
        response = self.get(url)
        return self.object_or_data(BatchErrors, response.json())

    def get_parameters(self, interface: str = None):
        """Make a GET request to the parameters endpoint"""
        url = self.urls.parameters()
        queryparams = None
        if interface:
            queryparams = {"interface": interface}
        response = self.get(url, params=queryparams)
        return self.object_or_data(Parameter, response.json())

    # Order ("Sotra") functions:

    def get_order(self, order_id: str) -> Union[Order, dict]:
        """
        GETs one order object
        """
        url = self.urls.order(str(order_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            return Order(**data)
        else:
            return data

    def get_detail(self, detail_id: str) -> Union[Detail, dict]:
        """
        GETs one detail object
        """
        url = self.urls.detail(str(detail_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            return Detail(**data)
        else:
            return data

    def get_order_list(self):
        """
        GETs a list of all orders, without detail objects
        """
        url = self.urls.order()
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        return self.object_or_data(Order, response.json())

    def get_details_in_order(self, order_id: str) -> Union[List[Detail], dict]:
        """
        GETs list of all detail objects belonging to an order
        """
        url = self.urls.details_in_order(str(order_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            if isinstance(data, list):
                return [Detail(**item) for item in data]
            elif isinstance(data, dict):
                return [Detail(**data)]
        else:
            return data

    def get_order_complete(self, order_id: str) -> Union[Order, dict]:
        """
        GETs one order, with all detail objects
        """
        url = self.urls.order_complete(str(order_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            return Order(**data)
        else:
            return data

    def get_abw_order_complete(self, abw_order_id: str) -> Union[AbwOrder, dict]:
        """
        GETs one abworder, with all order and detail objects
        """
        url = self.urls.abw_order_complete(str(abw_order_id))
        response = self.get(url)
        data = response.json()
        response.raise_for_status()
        if self.return_objects:
            return AbwOrder(**data)
        else:
            return data

    def post_add_abw_order(self, abworder: AbwOrder):
        """
        POST one AbwOrder, with its orders and its details.

        Note: There are some fields that cannot be sent in when creating a new AbwOrder:
        AbwOrder.id
        AbwOrder.progress
        AbwOrder.abworder_validated_ok_date
        Order.id
        Order.abw_order_id
        Detail.id
        Detail.order_id

        Returns tuple, with (data, status)
        """
        url = self.urls.add_abw_order()
        headers = {'Content-Type': 'application/json'}
        response = self.post(url,
                             data=abworder.json(),
                             headers=headers,
                             return_response=True)
        if response.status_code == 202:
            return response.content, 'Accepted'
        elif response.status_code == 409:
            return response.content, "Conflict"
        else:
            response.raise_for_status()
            return response

    def get_abw_order_errors(self, abw_order_id: str):
        """
        Gets an object containing three lists of all errors, for each of AbwOrder, Orders and Details
        """
        url = self.urls.abw_order_errors(abw_order_id)
        response = self.get(url)
        return self.object_or_data(AbwOrderErrors, response.json())


def get_client(config_dict):
    """
    Get a SetraClient from configuration.
    """
    return SetraClient(**config_dict)
